#!/usr/bin/env python3
"""
This is a support module for parsing [profile](https://www.w3.org/TR/dx-prof/)
metadata and applying entailment, validation and annotation operations
to RDF graphs.

Conformance to a given profile is declared by using
[`dcterms:conformsTo`](https://www.dublincore.org/specifications/dublin-core/dcmi-terms/#http://purl.org/dc/terms/conformsTo).

This module uses the following [resource roles](https://www.w3.org/TR/dx-prof/#Class:ResourceRole)
(where the `profrole` prefix is [http://www.w3.org/ns/dx/prof/role/](http://www.w3.org/ns/dx/prof/role/)):

* `profrole:entailment` for entailment operations (needs to conform to SHACL).
* `profrole:entailment-closure` is used as extra ontological information for entailment.
* `profrole:validation` for validation operations (needs to conform to SHACL).
* `profrole:validation-closure` is used as extra ontological information for validation.
* `profrole:annotation` is loaded as additional ontological annotation data.

"""
from __future__ import annotations
import itertools
import logging
from collections import deque
from typing import Union, Sequence, Optional, Generator, Any, cast, Iterable
from rdflib import Graph, RDF, PROF, OWL, URIRef, DCTERMS, Namespace, RDFS

from ogc.na import util
from pathlib import Path

from ogc.na.validation import ProfileValidationReport, ProfilesValidationReport, ValidationReport

PROFROLE = Namespace('http://www.w3.org/ns/dx/prof/role/')

ROLE_ENTAILMENT = PROFROLE.entailment
ROLE_ENTAILMENT_CLOSURE = PROFROLE['entailment-closure']
ROLE_VALIDATION = PROFROLE.validation
ROLE_VALIDATION_CLOSURE = PROFROLE['validation-closure']
ROLE_ANNOTATION = PROFROLE.annotation
ROLE_SEMANTIC_UPLIFT = PROFROLE['semantic-uplift']

PROFILES_QUERY = """
    PREFIX prof: <http://www.w3.org/ns/dx/prof/>
    PREFIX profrole: <http://www.w3.org/ns/dx/prof/role/>
    PREFIX shacl: <http://www.w3.org/ns/shacl#>
    PREFIX dct: <http://purl.org/dc/terms/>
    PREFIX owl: <http://www.w3.org/2002/07/owl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    CONSTRUCT {
        ?profile a prof:Profile ;
            prof:hasToken ?token ;
            prof:isProfileOf ?ancestor ;
            prof:hasResource ?resource ;
            owl:sameAs ?sameAs ;
            rdfs:label ?label ;
            .
        ?resource prof:hasRole ?role ;
            dct:conformsTo shacl: ;
            prof:hasArtifact ?artifact ;
            .
    } WHERE {
        __SERVICE__ {
            { ?profile a prof:Profile } UNION { ?other prof:isProfileOf ?profile }
            ?profile prof:hasToken ?token .
            OPTIONAL {
                { ?profile owl:sameAs+ ?sameAs } UNION { ?sameAs owl:sameAs+ ?profile }
            }
            OPTIONAL { ?profile rdfs:label ?label }
            OPTIONAL { ?profile prof:isProfileOf ?ancestor }
            OPTIONAL {
                ?profile prof:hasResource ?resource .
                ?resource prof:hasRole ?role ;
                    dct:conformsTo shacl: ;
                    prof:hasArtifact ?artifact
                OPTIONAL {
                    ?resource prof:hasRole ?role ;
                        prof:hasArtifact ?artifact
                }
            } 
        }
    }
"""

logger = logging.getLogger('ogc.na.profile')


def find_profiles(g: Graph) -> Generator[URIRef, Any, None]:
    return (o for s, o in g.subject_objects(DCTERMS.conformsTo) if isinstance(o, URIRef))


class ArtifactError(Exception):
    pass


class Profile:

    def __init__(self, uri: URIRef, token: str, profile_of: list[URIRef], label: str | None = None):
        self.uri = uri
        self.token = token
        self.label = label
        self.profile_of = [p for p in profile_of if p != uri]
        self.artifacts: dict[URIRef, list[str]] = {}

    def add_artifact(self, role: URIRef, href: URIRef):
        self.artifacts.setdefault(role, []).append(href)

    def get_artifacts(self, role: URIRef) -> list[URIRef]:
        return self.artifacts.get(role, [])

    def __repr__(self):
        return f"Profile({str(self.uri)},token={self.token}" \
               f",profileOf=[{','.join(str(p) for p in self.profile_of)}]" \
               f",roles=[{','.join(str(k) for k in self.artifacts.keys())}])"


class ProfileRegistry:

    def __init__(self, srcs: str | Path | Iterable[str | Path],
                 local_artifact_mappings: dict[str, str | Path] = None,
                 ignore_artifact_errors=False):

        assert srcs is not None
        if isinstance(srcs, str) or not isinstance(srcs, Iterable):
            self._srcs = (srcs,)
        else:
            self._srcs = srcs

        self._local_artifact_mappings: dict[str, Union[str, Path]] = {}
        if local_artifact_mappings:
            self._local_artifact_mappings = {u: Path(p) for u, p in local_artifact_mappings.items()}
        logger.debug("Using local artifact mappings: %s", self._local_artifact_mappings)
        self.profiles: dict[URIRef, Profile] = {}
        self._load_profiles()
        # Cache of { profile: { role: Graph } }
        self._graphs: dict[URIRef, dict[URIRef,
                           tuple[Graph, set[Union[str, Path]], set[Union[str, Path]]]]] = {}

        self.ignore_artifact_errors = ignore_artifact_errors

    def build_profile_chain(self, profiles: Sequence[URIRef | str],
                            recursive: bool = True,
                            sort: bool = True) -> list[URIRef]:
        if not profiles:
            return []

        if not sort:
            # Only known profiles and remove duplicates
            known = {URIRef(p): True for p in profiles if p in self.profiles}
            if recursive:
                pending = set(known)
                while pending:
                    prof = self.profiles.get(pending.pop())
                    if prof:
                        for prof_of in prof.profile_of or ():
                            if prof_of not in known:
                                pending.add(prof_of)
                            known[prof_of] = True
            return list(known)

        # Otherwise, sort DAG
        # 1. Build dependency tree
        dependencies: dict[URIRef, set[URIRef] | None] = {}
        pending = deque(profiles)  # using a deque to try and preserve as much of the original order as possible
        while pending:
            prof_uri = pending.popleft()
            if prof_uri in dependencies:
                # skip if already processed
                continue
            prof = self.profiles.get(prof_uri)
            if not prof:
                # skip if unknown
                continue
            prof_deps = [d for d in self.profiles.get(prof_uri).profile_of if d in self.profiles]
            if not prof_deps:
                # has no dependencies
                dependencies[prof_uri] = None
            if not recursive:
                # non-recursive => only work with provided profile list
                prof_deps = [d for d in prof_deps if d in profiles]
            else:
                for p in prof_deps:
                    if p not in pending:
                        pending.appendleft(p)
            dependencies[prof_uri] = set(prof_deps)

        result: list[URIRef] = []
        # 2. Sort dependencies
        while True:
            if not dependencies:
                break
            removed = {}
            for parent_uri, child_uris in dependencies.items():
                if not child_uris:
                    removed[parent_uri] = True
            if not removed:
                dependencies_str = '; '.join(
                    f"{p} <- {', '.join(str(c) for c in cs)}" for p, cs in dependencies.items())
                raise ValueError(f'Cycle detected in profile DAG, cannot sort: {dependencies_str}')
            for rem in removed:
                del dependencies[rem]
                result.append(rem)
            removed = set(removed)
            for child_uris in dependencies.values():
                child_uris -= removed

        return result

    def _load_profiles(self):
        logger.debug("Loading profiles from %s", [str(x) for x in self._srcs])
        g: Graph = Graph()
        for src in self._srcs:
            if isinstance(src, str) and src.startswith('sparql:'):
                endpoint = src[len('sparql:'):]
                logger.info("Fetching profiles from SPARQL endpoint %s", endpoint)
                assert util.is_url(endpoint, http_only=True)
                s = g.query(PROFILES_QUERY.replace('__SERVICE__', f"SERVICE <{endpoint}>")).graph
                util.copy_triples(s, g)
            else:
                g.parse(src)

        # resolve recursive isProfileOf and sameAs
        g = g.query(PROFILES_QUERY.replace('__SERVICE__', '')).graph

        for profile_ref in cast(list[URIRef], g.subjects(RDF.type, PROF.Profile)):

            if profile_ref in self.profiles:
                # do not parse duplicate profiles
                continue

            token = str(g.value(profile_ref, PROF.hasToken))
            label = g.value(profile_ref, RDFS.label)
            profile_of: list[URIRef] = cast(list[URIRef], list(g.objects(profile_ref, PROF.isProfileOf)))

            profile = Profile(profile_ref, token, profile_of, label=str(label) if label else None)

            for resource_ref in g.objects(profile_ref, PROF.hasResource):
                role_ref = g.value(resource_ref, PROF.hasRole)
                if not role_ref:
                    continue
                for artifact_ref in g.objects(resource_ref, PROF.hasArtifact):
                    profile.add_artifact(role_ref, cast(URIRef, artifact_ref))

            self.profiles[profile_ref] = profile
            for same_as_ref in g.objects(profile_ref, OWL.sameAs):
                self.profiles[cast(URIRef, same_as_ref)] = profile

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Profiles loaded: %s", [str(p) for p in self.profiles])
        else:
            logger.info(f"Loaded {len(self.profiles)}")

    def _apply_mappings(self, uri: str) -> Path | str:
        """
        Returns the longest match in self._local_artifact_mappings (prefixes)
        for a given URI, or the URI itself if not found
        """

        if uri in self._local_artifact_mappings:
            return self._local_artifact_mappings[uri]

        matchedlocal = None
        matchedpath = uri
        for l, p in self._local_artifact_mappings.items():
            if uri.startswith(l) and (matchedlocal is None or len(matchedlocal) < len(l)):
                matchedlocal, matchedpath = l, p / uri[len(l):]
        return matchedpath

    def get_artifacts(self, profile: URIRef | str, role: URIRef) -> set[str | Path] | None:
        if not isinstance(profile, URIRef):
            profile = URIRef(profile)
        if profile not in self.profiles:
            return None

        return set(self._apply_mappings(artifact_ref) for artifact_ref in self.profiles[profile].get_artifacts(role))

    def get_graph(self, profile: URIRef | str, role: URIRef) \
            -> tuple[Graph | None, set[str | Path] | None, set[str | Path] | None]:
        if not isinstance(profile, URIRef):
            profile = URIRef(profile)
        if profile not in self.profiles:
            return None, None, None

        prof_graphs = self._graphs.setdefault(profile, {})

        if role in prof_graphs:
            return prof_graphs[role]

        g = Graph()
        artifacts = set()
        failed_artifacts = set()
        for artifact in self.get_artifacts(profile, role):
            try:
                g.parse(artifact)
                artifacts.add(artifact)
            except Exception as e:
                if self.ignore_artifact_errors:
                    logger.warning("Error when retrieving or parsing artifact %s: %s",
                                   artifact, str(e))
                    failed_artifacts.add(artifact)
                else:
                    raise Exception(f"Error when retrieving or parsing artifact {artifact}") from e

            prof_graphs[role] = g, artifacts, failed_artifacts
        return g, artifacts, failed_artifacts

    def entail(self, g: Graph,
               additional_profiles: Optional[Sequence[str | URIRef]] = None,
               inplace: bool = True,
               recursive: bool = True) -> tuple[Graph, set[Union[str, Path]]]:
        if not inplace:
            g = util.copy_triples(g)

        profiles = deque(find_profiles(g))
        if additional_profiles:
            profiles.extend(p if isinstance(p, URIRef) else URIRef(p) for p in additional_profiles)

        profiles = self.build_profile_chain(profiles, recursive=recursive)

        artifacts = set()
        for profile_ref in profiles:
            logger.info('Entailing with %s', profile_ref)
            rules, rules_artifacts, failed_artifacts = self.get_graph(profile_ref, ROLE_ENTAILMENT)
            extra, extra_artifacts, failed_artifacts = self.get_graph(profile_ref, ROLE_ENTAILMENT_CLOSURE)
            if rules_artifacts:
                artifacts.update(rules_artifacts)
            if extra_artifacts:
                artifacts.update(extra_artifacts)
            g = util.entail(g, rules, extra or None, True)

        return g, artifacts

    def validate(self, g: Graph,
                 additional_profiles: Sequence[str | URIRef] | None = None,
                 recursive: bool = True, log_artifact_errors: bool = False) -> ProfilesValidationReport:
        result = ProfilesValidationReport()
        profiles = deque(find_profiles(g))
        if additional_profiles:
            profiles.extend(p if isinstance(p, URIRef) else URIRef(p) for p in additional_profiles)

        profiles = self.build_profile_chain(profiles, recursive=recursive, sort=False)

        for profile_ref in profiles:
            logger.info("Validating with %s", str(profile_ref))
            profile = self.profiles.get(profile_ref)
            if not profile:
                logger.warning("Profile %s not found", profile_ref)
                # should we fail?
                continue
            rules, rules_artifacts, failed_rules_artifacts = self.get_graph(profile_ref, ROLE_VALIDATION)
            extra, extra_artifacts, failed_extra_artifacts = self.get_graph(profile_ref, ROLE_VALIDATION_CLOSURE)
            failed_artifacts = failed_rules_artifacts | failed_extra_artifacts
            try:
                prof_result = util.validate(g, rules, extra)
            except Exception as e:
                if log_artifact_errors:
                    err_text = f"Error performing SHACL validation: {e}\n"
                    if rules_artifacts:
                        err_text += 'Used rules artifacts:\n' + '\n - '.join((str(s) for s in rules_artifacts))
                    if extra_artifacts:
                        err_text += 'Used extra artifacts:\n' + '\n - '.join((str(s) for s in extra_artifacts))
                    prof_result = ValidationReport((False, Graph(), err_text))
                else:
                    all_artifacts = {
                        'rules': [str(a) for a in rules_artifacts],
                        'extra': [str(a) for a in extra_artifacts],
                    }
                    raise ArtifactError('Error performing SHACL validation', all_artifacts) from e

            prof_result.used_resources = set(itertools.chain(rules_artifacts or [], extra_artifacts or []))
            if failed_artifacts:
                prof_result.text += "\n# Failed artifacts\n" + '\n'.join(str(a) for a in failed_artifacts) + '\n'
            result.add(ProfileValidationReport(profile_ref, profile.token, prof_result))
            logger.debug("Adding validation results for %s", profile_ref)

        return result

    def get_annotations(self, g: Graph, additional_profiles: Sequence[URIRef | str] | None = None) -> dict[Path, Graph]:
        result = {}
        profiles = find_profiles(g)
        if additional_profiles:
            profiles = itertools.chain(profiles, additional_profiles)
        for profile_ref in profiles:
            artifacts = self.get_artifacts(profile_ref, ROLE_ANNOTATION)
            for artifact in artifacts:
                result[artifact] = Graph().parse(artifact)
        return result

    def has_profile(self, uri: str | URIRef) -> bool:
        if isinstance(uri, str):
            uri = URIRef(uri)
        return uri in self.profiles



#!/usr/bin/env python3
from __future__ import annotations
import mimetypes
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Union, Optional, Sequence

from rdflib.term import Node

try:
    import git
except ImportError:
    git = None
from rdflib import Graph, URIRef, Literal, DCTERMS, RDF, RDFS, PROV, BNode

from ogc.na import __version__, __url__


@dataclass
class FileProvenanceMetadata:
    filename: Union[str, Path] = None
    uri: str = None
    mime_type: str = None
    use_bnode: bool = True
    label: str = None


@dataclass
class ProvenanceMetadata:
    used: Union[FileProvenanceMetadata, list[FileProvenanceMetadata]] = None
    generated: Union[FileProvenanceMetadata, list[FileProvenanceMetadata]] = None
    start: datetime = None
    end: datetime = None
    end_auto: bool = False
    root_directory: Union[str, Path] = None
    base_uri: str = None
    batch_activity_id: str = None
    activity_label: str = None
    comment: str = None


def add_provenance_agent(g: Graph, module_name: str = None) -> Node:
    agent = BNode()
    g.add((agent, RDFS.seeAlso, URIRef(__url__)))
    g.add((agent, RDF.type, PROV.Agent))
    g.add((agent, RDF.type, URIRef('https://schema.org/SoftwareApplication')))
    g.add((agent, RDFS.label, Literal("OGC-NA tools")))
    if module_name:
        g.add((agent, RDFS.comment, Literal("{} version {}".format(module_name, __version__))))
    g.add((agent, DCTERMS.hasVersion, Literal(__version__)))
    return agent


def add_provenance_entity(g: Graph, metadata: FileProvenanceMetadata = None,
                          root_directory: Optional[Union[str, Path]] = None,
                          base_uri: Optional[str] = None) -> URIRef:
    entity = None
    mime = None
    if metadata:
        mime = metadata.mime_type

        if metadata.uri:
            if metadata.use_bnode:
                entity = BNode()
                g.add((entity, RDFS.seeAlso, URIRef(metadata.uri)))
            else:
                entity = URIRef(metadata.uri)
        elif metadata.filename:
            filename = Path(metadata.filename).resolve()
            uri = None

            if not mime:
                try:
                    mime = mimetypes.guess_type(filename)[0]
                except:
                    pass

            if root_directory and base_uri:
                root_directory = Path(root_directory).resolve()
                if filename.is_relative_to(root_directory):
                    rel = filename.relative_to(root_directory)
                    uri = f"{base_uri}{'/' if '#' not in base_uri and not base_uri.endswith('/') else ''}{rel}"

            uri = uri if uri else filename.as_uri()
            if metadata.use_bnode:
                entity = BNode()
                g.add((entity, RDFS.seeAlso, URIRef(uri)))
            else:
                entity = URIRef(uri)

            if git:
                try:
                    git_repo = git.Repo(filename, search_parent_directories=True)
                    g.add((entity, DCTERMS.hasVersion, Literal(f"git:{git_repo.head.object.hexsha}")))
                except:
                    pass

    if not entity:
        entity = BNode()

    g.add((entity, RDF.type, PROV.Entity))

    if mime:
        g.add((entity, DCTERMS.format, Literal(mime)))

    if metadata.label:
        g.add((entity, RDFS.label, Literal(metadata.label)))

    return entity


def generate_provenance(g: Graph = None,
                        metadata: ProvenanceMetadata = None,
                        module_name: str = None) -> Graph:
    if g is None:
        g = Graph()

    if not metadata:
        metadata = ProvenanceMetadata()

    activity = BNode()

    agent = add_provenance_agent(g, module_name)

    if metadata.used:
        for used in metadata.used if isinstance(metadata.used, Sequence) else (metadata.used,):
            used = add_provenance_entity(g, metadata=used,
                                         root_directory=metadata.root_directory,
                                         base_uri=metadata.base_uri)
            g.add((activity, PROV.used, used))

    if metadata.generated:
        for generated in metadata.used if isinstance(metadata.generated, Sequence) else (metadata.generated,):
            generated = add_provenance_entity(g, metadata=generated,
                                              root_directory=metadata.root_directory,
                                              base_uri=metadata.base_uri)
            g.add((generated, PROV.wasGeneratedBy, activity))
            g.add((generated, PROV.wasAttributedTo, agent))

    g.add((activity, RDF.type, PROV.Activity))
    if metadata.activity_label:
        g.add((activity, RDFS.label, Literal(metadata.activity_label)))
    if metadata.start:
        g.add((activity, PROV.startedAtTime, Literal(metadata.start)))

    if metadata.end or metadata.end_auto:
        end = metadata.end if metadata.end else datetime.now()
        g.add((activity, PROV.endedAtTime, Literal(end)))
    if metadata.batch_activity_id:
        batch_activity = BNode()
        g.add((batch_activity, DCTERMS.identifier, Literal(metadata.batch_activity_id)))
        g.add((activity, PROV.wasInformedBy, batch_activity))
    if metadata.comment:
        g.add((activity, RDFS.comment, Literal(metadata.comment)))

    g.add((activity, PROV.wasAssociatedWith, agent))

    return g



#!/usr/bin/env python3
"""
General utilities module.
"""
from __future__ import annotations
import os.path
import shlex
from glob import glob
from pathlib import Path
from time import time
from typing import Optional, Union, Any, Mapping, Hashable

import requests
import rfc3987

from rdflib import Graph
from pyshacl import validate as shacl_validate
from urllib.parse import urlparse

from ogc.na.validation import ValidationReport

import yaml

try:
    from yaml import CLoader as YamlLoader, CSafeLoader as SafeYamlLoader, CDumper as YamlDumper
except ImportError:
    from yaml import Loader as YamlLoader, SafeLoader as SafeYamlLoader, Dumper as YamlDumper


def copy_triples(src: Graph, dst: Optional[Graph] = None) -> Graph:
    """
    Copies all triples from one graph onto another (or a new, empty [Graph][rdflib.Graph]
    if none is provided).

    :param src: the source Graph
    :param dst: the destination Graph (or `None` to create a new one)
    :return: the destination Graph
    """
    if dst is None:
        dst = Graph()
    for triple in src:
        dst.add(triple)
    return dst


def parse_resources(src: Union[str, Graph, list[Union[str, Graph]]]) -> Graph:
    """
    Join one or more RDF documents or [Graph][rdflib.Graph]'s together into
    a new Graph.
    :param src: a path or [Graph][rdflib.Graph], or list thereof
    :return: a union Graph
    """
    if not isinstance(src, list):
        src = [src]

    result = Graph()
    for s in src:
        if not isinstance(s, Graph):
            s = Graph().parse(s)
        copy_triples(s, result)

    return result


def entail(g: Graph,
           rules: Graph,
           extra: Optional[Graph] = None,
           inplace: bool = True) -> Graph:
    """
    Performs SHACL entailments on a data [Graph][rdflib.Graph].

    :param g: input data Graph
    :param rules: SHACL Graph for entailments
    :param extra: Graph with additional ontological information for entailment
    :param inplace: if `True`, the source Graph will be modified, otherwise a new
           Graph will be created
    :return: the resulting Graph
    """
    entailed_extra = None
    if extra:
        entailed_extra = copy_triples(extra)
        shacl_validate(entailed_extra, shacl_graph=rules, ont_graph=None, advanced=True, inplace=True)

    if not inplace:
        g = copy_triples(g)
    shacl_validate(g, shacl_graph=rules, ont_graph=extra, advanced=True, inplace=True)

    if entailed_extra:
        for triple in entailed_extra:
            g.remove(triple)

    return g


def validate(g: Graph, shacl_graph: Graph, extra: Optional[Graph] = None) -> ValidationReport:
    """
    Perform SHACL validation on a data [Graph][rdflib.Graph].

    :param g: input data Graph
    :param shacl_graph: SHACL graph for validation
    :param extra: Graph with additional ontological information for validation
    :return: the resulting [][ogc.na.validation.ValidationReport]
    """
    return ValidationReport(shacl_validate(data_graph=g,
                                           shacl_graph=shacl_graph,
                                           ont_graph=extra,
                                           inference='rdfs',
                                           advanced=True))


def is_url(url: str, http_only: bool = False) -> bool:
    """
    Checks whether a string is a valid URL.

    :param url: the input string
    :param http_only: whether to only accept HTTP and HTTPS URLs as valid
    :return: `True` if this is a valid URL, otherwise `False`
    """
    if not url:
        return False

    parsed = urlparse(url)
    if not parsed.scheme or not (parsed.netloc or parsed.path):
        return False

    if http_only and parsed.scheme not in ('http', 'https'):
        return False

    return True


def load_yaml(filename: str | Path | None = None,
              content: Any | None = None,
              url: str | None = None,
              safe: bool = True) -> dict:
    """
    Loads a YAML file either from a file, a string or a URL.

    :param filename: YAML document file name
    :param content: str with YAML contents
    :param url: url from which to retrieve the contents
    :param safe: whether to use safe YAMl loading
    :return: a dict with the loaded data
    """

    if bool(filename) + bool(content) + bool(url) > 1:
        raise ValueError("One (and only one) of filename, contents and url must be provided")

    if filename:
        with open(filename, 'r') as f:
            return yaml.load(f, Loader=SafeYamlLoader if safe else YamlLoader)
    else:
        if url:
            content = requests.get(url).text
        return yaml.load(content, Loader=SafeYamlLoader if safe else YamlLoader)


def dump_yaml(content: Any, filename: str | Path | None = None,
              **kwargs) -> str | None:
    """
    Generates YAML output.

    :param content: content to convert to YAML.
    :param filename: optional filename to dump the content into. If None, string content will be returned.
    :param kwargs: other args to pass to `yaml.dump()`
    """
    kwargs.setdefault('sort_keys', False)
    if filename:
        with open(filename, 'w') as f:
            return yaml.dump(content, f, Dumper=YamlDumper, **kwargs)
    else:
        return yaml.dump(content, Dumper=YamlDumper, **kwargs)


def is_iri(s: str) -> bool:
    try:
        return rfc3987.parse(s, rule='IRI') is not None
    except ValueError:
        return False


def merge_dicts(src: dict, dst: dict) -> dict:
    if not src:
        return dst
    for k, v in src.items():
        if isinstance(v, dict):
            node = dst.setdefault(k, {})
            merge_dicts(v, node)
        else:
            dst[k] = v
    return dst


def glob_list_split(s: str, exclude_dirs: bool = True, recursive: bool = False) -> list[str]:
    result = []
    for e in shlex.split(s):
        if is_url(s):
            result.append(s)
        else:
            for fn in glob(e, recursive=recursive):
                if not exclude_dirs or os.path.isfile(fn):
                    result.append(fn)
    return result


class LRUCache:

    def __init__(self, maxsize: int = 10):
        self._cache: dict[Hashable, Any] = {}
        self._last_access: dict[Hashable, float] = {}
        self._maxsize = maxsize

    def __contains__(self, item):
        return item in self._cache

    def __len__(self):
        return len(self._cache)

    def get(self, key, default=None):
        if not isinstance(key, Hashable):
            return default
        return self._cache.get(key, default)

    def __setitem__(self, key, value):
        if not isinstance(key, Hashable):
            return
        if len(self._cache) >= self._maxsize and key not in self._cache:
            key_to_remove = min(self._last_access, key=self._last_access.get)
            del self._cache[key_to_remove]
            del self._last_access[key_to_remove]
        self._cache[key] = value
        self._last_access[key] = time()


def deep_update(orig_dict: dict, with_dict: Mapping, replace: bool = False) -> dict:
    if not isinstance(orig_dict, Mapping):
        return with_dict
    dest = orig_dict if replace else {**orig_dict}
    for k, v in with_dict.items():
        if isinstance(v, Mapping):
            dest[k] = deep_update(orig_dict.get(k, {}), v, replace)
        else:
            dest[k] = v
    return dest


def git_status(repo_path: str | Path = '.'):
    import git
    repo = git.Repo(repo_path)
    added = repo.untracked_files
    modified = []
    deleted = []
    renamed = []
    for diff in repo.head.commit.diff(None):
        if diff.change_type == 'D':
            deleted.append(diff.a_path)
        elif diff.change_type == 'M':
            modified.append(diff.a_path)
        elif diff.change_type == 'R':
            renamed.append((diff.a_path, diff.b_path))
    return {
        'added': added,
        'modified': modified,
        'deleted': deleted,
        'renamed': renamed,
    }
